/** Public (re-exported by alleycat.h)
 */

#ifndef _AC_MEM_MACROS_H
#define _AC_MEM_MACROS_H

#include "mem.h"

/* Usage:
 *
 * On error, the _rf variants return false, the _rnull variants return null and the _die variants
 * cause the program to exit. Not all variants are shown in the usage.
 *
 * Don't use any of these macros in an if without braces: some of them will work, but the ones which
 * declare a variable can't be wrapped in 'do {} while (0);' and will cause annoying bugs.
 *
 *   try_malloc (struct state, state)
 *   // struct state *state = malloc (sizeof (struct state));
 *   if (state == NULL) ...
 *
 *   try_malloc (struct state *, state)
 *   // struct state **state = malloc (sizeof (struct state *));
 *   if (state == NULL) ...
 *
 *   try_malloc_rf (struct state, state)
 *   // struct state *state = malloc (sizeof (struct state));
 *   // if (state == NULL) return false;
 *
 *   struct state *state;
 *   try_malloc_size_nodecl_rf (state, sizeof (struct state))
 *   // state = malloc (sizeof (struct state));
 *
 *   try_malloc_size (struct state, state, some_size)
 *   // struct state *state = malloc (some_size);
 *   if (state == NULL) ...
 *
 *   try_malloc_size (struct state *, state, some_size)
 *   // struct state **state = malloc (some_size);
 *   if (state == NULL) ...
 *
 *   try_malloc_size (const struct state, state, some_size)
 *   // const struct state *state = malloc (some_size);
 *   if (state == NULL) ...
 *
 *   try_malloc_size (const struct state, const state, some_size)
 *   // const struct state * const state = malloc (some_size);
 *   if (state == NULL) ...
 *
 *   try_calloc_rf (struct state *, state, 10)
 *   // struct state **state = calloc (10, sizeof (struct state *));
 *
 *   try_realloc_rf (struct state, state_new, state)
 *   // struct state *state_new = realloc (state, sizeof (struct state));
 *
 *   try_realloc_size_nodecl_rf (state_new, state, some_size)
 *   // state_new = realloc (state, some_size);
 */

#define try_malloc_size(type, var, size) \
  type *var = ac_malloc_non_fatal (size);

#define try_malloc_size_rnull(type, var, size) \
  type *var = ac_malloc_non_fatal (size); \
  if (var == NULL) return NULL;

#define try_malloc_size_rf(type, var, size) \
  type *var = ac_malloc_non_fatal (size); \
  if (var == NULL) return false;

#define try_malloc_size_die(type, var, size) \
  type *var = ac_malloc (size);

#define try_malloc(type, var) \
  try_malloc_size (type, var, sizeof(type));

#define try_malloc_rnull(type, var) \
  try_malloc_size_rnull (type, var, sizeof(type));

#define try_malloc_rf(type, var) \
  try_malloc_size_rf (type, var, sizeof(type));

#define try_malloc_die(type, var) \
  try_malloc_size_die (type, var, sizeof(type));

#define try_malloc_size_nodecl(var, size) do { \
  var = ac_malloc_non_fatal (size); \
} while (0);

#define try_malloc_size_nodecl_rnull(var, size) do { \
  var = ac_malloc_non_fatal (size); \
  if (var == NULL) return NULL; \
} while (0);

#define try_malloc_size_nodecl_rf(var, size) do { \
  var = ac_malloc_non_fatal (size); \
  if (var == NULL) return false; \
} while (0);

#define try_malloc_size_nodecl_die(var, size) do { \
  var = ac_malloc (size); \
} while (0);


#define try_calloc_size(type, var, nmemb, size) \
  type *var = ac_calloc_non_fatal (nmemb, size);

#define try_calloc_size_rnull(type, var, nmemb, size) \
  type *var = ac_calloc_non_fatal (nmemb, size); \
  if (var == NULL) return NULL;

#define try_calloc_size_rf(type, var, nmemb, size) \
  type *var = ac_calloc_non_fatal (nmemb, size); \
  if (var == NULL) return false;

#define try_calloc_size_die(type, var, nmemb, size) \
  type *var = ac_calloc (nmemb, size);

#define try_calloc(type, var, nmemb) \
  try_calloc_size (type, var, nmemb, sizeof(type));

#define try_calloc_rnull(type, var, nmemb) \
  try_calloc_size_rnull (type, var, nmemb, sizeof(type));

#define try_calloc_rf(type, var, nmemb) \
  try_calloc_size_rf (type, var, nmemb, sizeof(type));

#define try_calloc_die(type, var, nmemb) \
  try_calloc_size_die (type, var, nmemb, sizeof(type));

#define try_calloc_size_nodecl(var, nmemb, size) do { \
  var = ac_calloc_non_fatal (nmemb, size); \
} while (0);

#define try_calloc_size_nodecl_rnull(var, nmemb, size) do { \
  var = ac_calloc_non_fatal (nmemb, size); \
  if (var == NULL) return NULL; \
} while (0);

#define try_calloc_size_nodecl_rf(var, nmemb, size) do { \
  var = ac_calloc_non_fatal (nmemb, size); \
  if (var == NULL) return false; \
} while (0);

#define try_calloc_size_nodecl_die(var, nmemb, size) do { \
  var = ac_calloc (nmemb, size); \
} while (0);


#define try_realloc_size(type, var, ptr, size) \
  type *var = ac_realloc_non_fatal (ptr, size);

#define try_realloc_size_rnull(type, var, ptr, size) \
  type *var = ac_realloc_non_fatal (ptr, size); \
  if (var == NULL) return NULL;

#define try_realloc_size_rf(type, var, ptr, size) \
  type *var = ac_realloc_non_fatal (ptr, size); \
  if (var == NULL) return false;

#define try_realloc_size_die(type, var, ptr, size) \
  type *var = ac_realloc (ptr, size);

#define try_realloc(type, var, ptr) \
  try_realloc_size (type, var, ptr, sizeof(type));

#define try_realloc_rnull(type, var, ptr) \
  try_realloc_size_rnull (type, var, ptr, sizeof(type));

#define try_realloc_rf(type, var, ptr) \
  try_realloc_size_rf (type, var, ptr, sizeof(type));

#define try_realloc_die(type, var, ptr) \
  try_realloc_size_die (type, var, ptr, sizeof(type));

#define try_realloc_size_nodecl(var, ptr, size) do { \
  var = ac_realloc_non_fatal (ptr, size); \
} while (0);

#define try_realloc_size_nodecl_rnull(var, ptr, size) do { \
  var = ac_realloc_non_fatal (ptr, size); \
  if (var == NULL) return NULL; \
} while (0);

#define try_realloc_size_nodecl_rf(var, ptr, size) do { \
  var = ac_realloc_non_fatal (ptr, size); \
  if (var == NULL) return false; \
} while (0);

#define try_realloc_size_nodecl_die(var, ptr, size) do { \
  var = ac_realloc (ptr, size); \
} while (0);





#endif
