/** Public (re-exported by alleycat.h)
 */

#ifndef _AC_SQUEAK_MACROS_H
#define _AC_SQUEAK_MACROS_H

#include <stdlib.h>

/*
 * iwarn, ierr: intended for programmer (not user) errors.
 *
 * All err functions exit with status of 1.
 *
 * ac_iwarn("")/ac_piep:    <file>:<line> Something's wrong (internally).
 * ac_iwarn(msg):      <file>:<line> Internal warning: <msg>
 * ac_ierr("")/ac_die:      <file>:<line> Internal error.
 * ac_ierr(msg):       <file>:<line> Internal error: <msg>
 *
 * ac_iwarn_perr(""):    <file>:<line> Something's wrong (internally) (<system error>).
 * ac_iwarn_perr(msg): <file>:<line> Internal warning: <msg> (<system error>).
 * ac_ierr_perr(""):     <file>:<line> Internal error (<system error>).
 * ac_ierr_perr(msg):  <file>:<line> Internal error: <msg> (<system error>).
 *
 * ac_err, ac_warn: intended for user errors and system errors.
 *
 * ac_warn(""):          Something's wrong.
 * ac_warn(msg):       <msg>
 * ac_err(""):           Error.
 * ac_err(msg):        Error: <msg>
 *
 * ac_warn_perr(""):     Something's wrong (<system error>).
 * ac_warn_perr(msg):  <msg> (<system error>).
 * ac_err_perr(""):      Error (<system error>).
 * ac_err_perr(msg):   Error: <msg> (<system error>).
 *
 * ac_warn_aserr_xxx(...)  [to make it look like an error, but not die]
 *
 * ac_piep, ac_pieprf, piepr0, etc.: (pronounced 'peep').
 *
 * Intended as easy-to-type way to ac_warn (and possibly return false, return
 * 0, etc.). Pipes through iwarn().
 *
 * ac_die: synonym for ac_ierr()
 *
 * The error functions will often hold a few open pointers upon program
 * exit. It's too difficult (and overkill) to try to free them before exit
 * (they could be static).
 *
 */

#define ac_iwarn(format...) do { \
    ac_complain(__FILE__, __LINE__, false, false, format); \
} while (0)

#define ac_iwarn_aserr(format...) do { \
    ac_complain(__FILE__, __LINE__, true, false, format); \
} while (0)

#define ac_ierr(format...) do { \
    ac_complain(__FILE__, __LINE__, true, false, format); \
    exit(1); \
} while (0)

#define ac_die do { \
    ac_ierr(); \
} while (0)

#define ac_iwarn_perr(format...) do { \
    ac_complain(__FILE__, __LINE__, false, true, format); \
} while (0)

#define ac_iwarn_aserr_perr(format...) do { \
    ac_complain(__FILE__, __LINE__, true, true, format); \
} while (0)

#define ac_ierr_perr(format...) do { \
    ac_complain(__FILE__, __LINE__, true, true, format); \
    exit(1); \
} while (0)

#define ac_warn(format...) do { \
    ac_complain("", 0, false, false, format); \
} while (0)

#define ac_warn_aserr(format...) do { \
    ac_complain("", 0, true, false, format); \
} while (0)

#define ac_warn_rf(args...) do { \
  ac_warn(args); \
  return false; \
} while (0);

#define ac_warn_rnull(args...) do { \
  ac_warn(args); \
  return NULL; \
} while (0);

#define ac_warn_perr_rf(args...) do { \
  ac_warn_perr(args); \
  return false; \
} while (0);

#define ac_warn_perr_rnull(args...) do { \
  ac_warn_perr(args); \
  return NULL; \
} while (0);

#define ac_err(format...) do { \
    ac_complain("", 0, true, false, format); \
    exit(1); \
} while (0)

#define ac_warn_perr(format...) do { \
    ac_complain("", 0, false, true, format); \
} while (0)

#define ac_warn_aserr_perr(format...) do { \
    ac_complain("", 0, true, true, format); \
} while (0)

#define ac_err_perr(format...) do { \
    ac_complain("", 0, true, true, format); \
    exit(1); \
} while (0)

#define ac_piep do { \
    ac_iwarn(""); \
} while (0)

#define ac_pieprf do { \
    ac_piep; \
    return false; \
} while (0)

#define ac_pieprt do { \
    ac_piep; \
    return true; \
} while (0)

#define ac_piepr1 do { \
    ac_piep; \
    return 1; \
} while (0)

#define ac_piepr do { \
    ac_piep; \
    return; \
} while (0)

/* Don't use this in an if without braces.
 */
#define ac_piepbr \
    ac_piep; \
    break;

/* Don't use this in an if without braces.
 */
#define ac_piepc \
    ac_piep; \
    continue;

#define ac_pieprnull do { \
    ac_piep; \
    return NULL; \
} while (0)

#define ac_pieprneg1 do { \
    ac_piep; \
    return -1; \
} while (0)

#define ac_piepr0 do { \
    ac_piep; \
    return 0; \
} while (0)

#endif
