#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include <time.h>

#include <alleycat.h>

#include "private/crc.h"
#include "private/ds1961.h"
#include "private/ds2482.h"
#include "private/i2c.h"
#include "private/types.h"
#include "private/util.h"

#define ALLOW_0x4A 1
#define EXTRA_WAIT_PROG 1
#define NUM_RETRIES_WRITE_SECRET 10

static bool command (const struct key * const key, uint8_t const command) {
  const struct master * const master = key->bus->master;
  try (ds2482_key_command_init_skip_rom, (master))
  try (ds2482_write_byte, (master, command))
  return true;
}

bool ds1961_write_scratchpad (const struct key * const key, uint16_t addr, uint8_t * const data) {
  const struct master * const master = key->bus->master;
  try (command, (key, CMD_WRITE_SCRATCHPAD))

  uint8_t write[10];
  write[0] = (addr >> 0) & 0xFF;
  write[1] = (addr >> 8) & 0xFF;
  memcpy (write + 2, data, 8);
  try (ds2482_write_bytes, (master, write, 10))

  uint8_t data_crc[13];
  data_crc[0] = CMD_WRITE_SCRATCHPAD;
  memcpy (data_crc + 1, write, 10);
  try (ds2482_read_bytes, (master, data_crc + 11, 2))
  try_with (ds1961_inv_crc16_validate, (data_crc, 13), ac_warn_rf)
  return true;
}

/**
 * `data`: uint8_t[8]
 */
static bool read_scratchpad (const struct key * const key, uint16_t * const addr, uint8_t * const es, uint8_t * const read) {
  const struct master * const master = key->bus->master;
  try (command, (key, CMD_READ_SCRATCHPAD))

  uint8_t data_read[3];
  try (ds2482_read_bytes, (master, data_read, 3))
  *addr = (data_read[1] << 8) | data_read[0];
  *es = data_read[2];

  try (ds2482_read_bytes, (master, read, 8))

  uint8_t data_crc[14];
  data_crc[0] = CMD_READ_SCRATCHPAD;
  memcpy (data_crc + 1, data_read, 3);
  memcpy (data_crc + 4, read, 8);
  try (ds2482_read_bytes, (master, data_crc + 12, 2))

  try_with (ds1961_inv_crc16_validate, (data_crc, 14), ac_warn_rf)
  return true;
}

static bool load_first_secret (const struct key * const key, uint16_t addr, uint8_t es) {
  const struct master * const master = key->bus->master;
  uint8_t write[3] = {
    (addr >> 0) & 0xFF,
    (addr >> 8) & 0xFF,
    es,
  };

  try (command, (key, CMD_LOAD_FIRST_SECRET))
  try (ds2482_write_bytes, (master, write, 3))

  nanosleep (&T_PROG, NULL);
  if (EXTRA_WAIT_PROG) nanosleep (&T_PROG, NULL);
  uint8_t status;
  try (i2c_read_byte, (master, &status))
  if (status == 0x4a) {
    if (ALLOW_0x4A) {
      ac_info ("load_first_secret () returned 0x4a, allowing");
      return true;
    }
    ac_info ("load_first_secret () returned 0x4a, not allowing");
    return false;
  }
  return status == 0xAA;
}

bool ds1961_inv_crc16_validate (const uint8_t * const data, uint8_t const length) {
  uint8_t byte_count = length - 2;
  uint8_t byte1_crc_expected = length - 2;
  uint8_t byte2_crc_expected = length - 1;
  return crc_validate16 (
    data, byte_count,
    ~data[byte1_crc_expected],
    ~data[byte2_crc_expected]
  );
}

/* Can fail due to CRC error, which doesn't set errno.
 */
bool ds1961_read_auth_page (const struct key * const key, uint8_t * const rmac_read, uint8_t * const rmac) {
  uint16_t auth_addr = key->auth_addr;
  const struct master * const master = key->bus->master;
  try (command, (key, CMD_READ_AUTH_PAGE))

  uint8_t write[2];
  write[0] = (auth_addr >> 0) & 0xFF;
  write[1] = (auth_addr >> 8) & 0xFF;
  try (ds2482_write_bytes, (master, write, 2))

  uint8_t read1[35];
  try (ds2482_read_bytes, (master, read1, 35))
  // @todo warning?
  if (read1[32] != 0xFF) return false;

  uint8_t data_crc[38];
  data_crc[0] = CMD_READ_AUTH_PAGE;
  memcpy (data_crc + 1, write, 2);
  memcpy (data_crc + 3, read1, 35);
  try_with_info (ds1961_inv_crc16_validate, (data_crc, 38), ac_warn_rf, (pass 1))

  memcpy (rmac_read, read1, 32);

  nanosleep (&T_CSHA, NULL);
  uint8_t read_mac[20];
  try (ds2482_read_bytes, (master, read_mac, 20))
  memcpy (rmac, read_mac, 20);

  uint8_t data_crc2[22];
  memcpy (data_crc2, read_mac, 20);
  try (ds2482_read_bytes, (master, data_crc2 + 20, 2))
  try_with_info (ds1961_inv_crc16_validate, (data_crc2, 22), ac_warn_rf, (pass 2))

  uint8_t status;
  try (ds2482_read_byte, (master, &status))
  return status == 0xAA;
}

static bool write_secret (const struct key * const key, const uint8_t * const secret16) {
  uint8_t secret8[8];
  secret_16_to_8 (secret16, secret8);
  try (ds1961_write_scratchpad, (key, MEM_REGION_SECRET, secret8))

  uint16_t addr;
  uint8_t es;
  uint8_t read[8];
  try (read_scratchpad, (key, &addr, &es, read))
  try_with (load_first_secret, (key, addr, es), ac_warn_rf)

  return true;
}

bool ds1961_generate_secret (const struct key * const key, uint8_t * const secret) {
  // --- 10ms
  static const struct timespec pass_wait = { .tv_sec = 0, .tv_nsec = 10e6, };

  srandom (time (NULL));
  for (int i = 0; i < 16; i++)
    secret[i] = (rand () % (0x7F -1)) + 1;
  for (int loop = 0; loop < NUM_RETRIES_WRITE_SECRET; loop++) {
    if (write_secret (key, secret))
      return true;
    ac_info ("write_secret (): failed (pass %d/%d)", loop + 1, NUM_RETRIES_WRITE_SECRET);
    nanosleep (&pass_wait, NULL);
  }
  ac_warn ("ds1961_generate_secret (): failed at write_secret ()");
  return false;
}
