#include <fcntl.h>
#include <stdint.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <alleycat.h>

#include <i2c-dev-linux.h>

#include "drup/types.h"
#include "private/auth.h"
#include "private/bus.h"
#include "private/ds1961.h"
#include "private/ds2482.h"
#include "private/i2c.h"
#include "private/search.h"
#include "private/types.h"
#include "private/util.h"

struct drup_state *drup_init (const struct drup_config * const config) {
  const struct drup_device * const device = config->device;

  const char * const i2c_dev = device->i2c_dev;
  int const address = device->address;
  int const ds2482_type = device->subtype;

  int const num_buses = ds2482_num_buses (ds2482_type);
  if (num_buses == -1) ac_warn_rnull (
    "init (): invalid ds2482_type: %d", ds2482_type
  );

  int const fd_master = open (i2c_dev, O_RDWR);
  if (fd_master == -1) ac_warn_perr_rnull (
    "Couldn't open device %s", i2c_dev
  );
  try_malloc (struct drup_state, state)
  if (state == NULL) ac_pieprnull;
  try_malloc (struct master, master)
  if (master == NULL) ac_pieprnull;
  try_calloc (struct bus *, buses, num_buses)
  master->fd = fd_master;

  for (int i = 0; i < num_buses; i++)
    buses[i] = bus_new (master, i);

  master->num_buses = num_buses;
  master->buses = buses;
  master->selected_bus = -1;

  state->config = (struct drup_config *) config;
  state->master = master;

  if (ioctl (fd_master, I2C_SLAVE, address)) {
    ac_warn_perr ("ioctl failed");
    return NULL;
  }
  debug ("resetting");
  if (!i2c_cmd (master, CMD_RESET)) ac_warn_perr_rnull (
    "Master not responding (invalid address 0x%x?)", address
  )
  if (config->verbose) ac_info ("initted");
  return state;
}

static void drup_state_destroy (struct drup_state * const state) {
  close (state->master->fd);
  for (int i = 0; i < state->master->num_buses; i++) {
    free (state->master->buses[i]);
  }
  free (state->master->buses);
  free (state->master);
  free (state);
}

void drup_cleanup (struct drup_state * const state) {
  if (state) drup_state_destroy (state);
}

static bool validate_bus_number (const struct drup_state * state, uint8_t const bus_number) {
  const struct master * const master = state->master;
  int const num_buses = master->num_buses;
  if (bus_number >= num_buses) ac_warn_rf (
    "%s (): invalid bus number %d", __func__, bus_number
  );
  const struct bus * const bus = master->buses[bus_number];
  if (bus == NULL) ac_warn_rf (
    "%s (): no bus found at number %d", __func__, bus_number
  );
  return true;
}

bool drup_search_init (const struct drup_state * const state, uint8_t bus_number, uint8_t cfg, bool * const presence) {
  if (get_bit (cfg, CFG_REG_SPU)) {
    ac_warn ("drup_search_init (): strong pullup is not implemented, setting to false");
    cfg &= ~CFG_REG_SPU;
  }
  try_with (validate_bus_number, (state, bus_number), ac_warn_rf)
  const struct bus * const bus = state->master->buses[bus_number];
  try (search_init, (state, bus, cfg, presence))
  return true;
}

bool drup_search (const struct drup_state * const state, uint8_t const device_family, uint64_t * const device_id, char * const device_id_str) {
  *device_id = 0;
  try (search_start, (state, device_family, device_id))
  if (*device_id) uint64_to_hex (*device_id, device_id_str);
  return true;
}

bool drup_authenticate (const struct drup_state * const state, uint8_t const bus_number, uint8_t cfg, uint64_t const device_id, const uint8_t * const secret, bool * const authenticated) {
  if (get_bit (cfg, CFG_REG_SPU)) {
    ac_warn ("drup_authenticate (): strong pullup is not implemented, setting to false");
    cfg &= ~CFG_REG_SPU;
  }
  try_with (validate_bus_number, (state, bus_number), ac_warn_rf)
  const struct master * const master = state->master;
  const struct bus * const bus = master->buses[bus_number];
  struct key key = {
    .bus = bus,
    .device_id = device_id,
    .secret = secret,
    // --- always 0.
    .auth_addr = 0,
  };
  return auth_check (state, &key, cfg, authenticated);
}

bool drup_generate_secret (const struct drup_state * const state, uint8_t const bus_number, uint8_t cfg, uint64_t const device_id, uint8_t * const secret) {
  // --- @todo: we don't currently do anything with cfg, since we don't do
  // any ds2482 commands here.
  (void) cfg;

  const struct master * const master = state->master;
  const struct bus * const bus = master->buses[bus_number];
  struct key key = {
    .bus = bus,
    .device_id = device_id,
    .secret = NULL,
    // --- always 0.
    .auth_addr = 0,
  };
  try (ds1961_generate_secret, (&key, secret))
  return true;
}

uint8_t drup_cfg (bool overdrive, bool strong_pullup, bool active_pullup) {
  return 0
    | (overdrive ? CFG_REG_1WS : 0)
    | (strong_pullup ? CFG_REG_SPU : 0)
    | (active_pullup ? CFG_REG_APU : 0);
}
